# SPDX-FileCopyrightText: 2021 Jose David
#
# SPDX-License-Identifier: MIT
"""
Creates a single color picker
"""

import time
import board
import terminalio
from displayio import Group, TileGrid, Bitmap, Palette
from adafruit_display_text import bitmap_label
import adafruit_touchscreen
from color_picker import color_picker


display = board.DISPLAY

# TouchScreen Configuration
ts = adafruit_touchscreen.Touchscreen(
    board.TOUCH_XL,
    board.TOUCH_XR,
    board.TOUCH_YD,
    board.TOUCH_YU,
    calibration=((5200, 59000), (5800, 57000)),
    size=(display.width, display.height),
)

# Colorwheel Bitmap file
filename = "wheel200.bmp"  # You can find this file in the examples directory in the library Github
# Change the imagesize_used according to the bitmap file used. Colorwheel are identified
# according to the size in pixels
imagesize_used = 200
my_colorpicker = color_picker.ColorPicker(
    filename,
    display.width // 2 - imagesize_used // 2,
    display.height // 2 - imagesize_used // 2,
    imagesize_used,
)
my_group = Group()
my_group.append(my_colorpicker)

palette = Palette(2)
palette[0] = 0x990099
palette[1] = 0x00FFFF

bitmap = Bitmap(100, 20, 2)
color_square = TileGrid(bitmap, pixel_shader=palette, x=display.width - 100, y=10)
my_group.append(color_square)
# Adding text information
text_area = bitmap_label.Label(
    terminalio.FONT,
    text="Color",
    x=display.width - 100,
    y=35,
)
my_group.append(text_area)

# Add my_group to the display
display.show(my_group)

p = False
# Start the main loop
while True:
    p = ts.touch_point
    if p:  # Check if colorpicker is selected
        if my_colorpicker.contains(p):
            color = my_colorpicker.when_selected(p, display.height)
            palette[0] = color
            print(f"Color Selected is: {hex(color)}")
            text_area.text = str(hex(color))
            time.sleep(1.5)

    time.sleep(0.05)  # touch response on PyPortal is more accurate with a small delay



# SPDX-FileCopyrightText: 2021 Jose David
#
# SPDX-License-Identifier: MIT
"""
Creates a bitmap color picker of 100 pixels
"""

from color_picker.wheel_maker import make_wheel

make_wheel("wheel100.bmp", 100, 0x000000)



# SPDX-FileCopyrightText: Copyright (c) 2021 Jose David M.
#
# SPDX-License-Identifier: MIT
"""
`color_picker`
================================================================================

CircuitPython helper library to select color using a color wheel in resistive
displays using DisplayIO library


* Author(s): Jose David M.

Implementation Notes
--------------------



"""
################################
# A color picker for CircuitPython, using displayio
#
# Features:
#  - color picker using a existing Bitmap
#
# Future options to consider:
# ---------------------------
# Better color wheel logic
# Sliders to mix Red, Green and Blue
#

import math
from displayio import TileGrid, OnDiskBitmap, ColorConverter
from adafruit_displayio_layout.widgets.widget import Widget
from adafruit_displayio_layout.widgets.control import Control


__version__ = "0.0.0-auto.0"
__repo__ = (
    "https://github.com/adafruit/Adafruit_CircuitPython_DisplayIO_Color_Picker.git"
)

# pylint: disable=missing-class-docstring, too-few-public-methods


class ColorPicker(Widget, Control):
    """A widget to be used to select colors from a heel.

    :param str filename: name of the bitmap file to be used as a ColorPicker

    :param int x: x position of the color picker origin
    :param int y: y position of the color picker origin

    :param int imagesize: size of the bitmap file. The bitmap colorwheels are squares.


    **Quickstart: Importing and using the Color Picker**

    Here is one way of importing the ``ColorPicker`` class so you can use:

    .. code-block:: python

        from adafruit_displayio_color_picker import color_picker

    Now you can create an Slider at pixel position x=20, y=30 using:

    .. code-block:: python

        my_colorpicker=color_picker.ColorPicker(x=20, y=30)

    Once you setup your display, you can now add ``my_colorpicker`` to your display using:

    .. code-block:: python

        display.show(my_colorpicker) # add the group to the display

    If you want to have multiple display elements, you can create a group and then
    append the slider and the other elements to the group.  Then, you can add the full
    group to the display as in this example:

    .. code-block:: python

        my_colorpicker= ColorPicker(20, 30)
        my_group = displayio.Group(max_size=2) # make a group that can hold 2 items
        my_group.append(my_colorpicker) # Add my_colorpicker to the group

        #
        # Append other display elements to the group
        #

        display.show(my_group) # add the group to the display


    **Final Notes**

    Depending on the screen results may vary. Resolution of the bitmap will no be as seen in a PC.
    Sensitivity of the screen could also affect the behaviour of the library.


    **The Color Picker Widget**

    .. figure:: color_picker.png
      :scale: 100 %
      :align: center
      :alt: Example of the color picker widget.

      Example of the color picker. representation will vary according to screen used.



    """

    # pylint: disable=too-many-lines, too-many-instance-attributes, too-many-arguments
    # pylint: disable=too-many-locals, too-many-statements, consider-using-with
    # pylint: disable=duplicate-code

    def __init__(
        self,
        filename: str = None,
        x: int = 0,
        y: int = 0,
        image_size: int = 100,
        **kwargs,
    ):

        Widget.__init__(self, x=x, y=y, height=image_size, width=image_size, **kwargs)
        Control.__init__(self)

        self._file = open(filename, "rb")
        image = OnDiskBitmap(self._file)
        tile_grid = TileGrid(image, pixel_shader=ColorConverter())

        self._image_size = image_size

        self.append(tile_grid)

        self.touch_boundary = (
            0,
            0,
            image.width,
            image.height,
        )

    def contains(self, touch_point):  # overrides, then calls Control.contains(x,y)

        """Checks if the ColorPicker was touched.  Returns True if the touch_point is
        within the ColorPicker's touch_boundary.

        :param touch_point: x,y location of the screen, converted to local coordinates.
        :type touch_point: Tuple[x,y]
        :return: Boolean
        """

        touch_x = (
            touch_point[0] - self.x
        )  # adjust touch position for the local position
        touch_y = touch_point[1] - self.y

        return super().contains((touch_x, touch_y, 0))

    def when_selected(self, touch_point, screen_height):
        """Response function when ColorPicker is selected.  When selected, the ColorPicker
        will give the color corresponding with the position

        :param touch_point: x,y location of the screen, in absolute display coordinates.
        :param int screen_height: screen height
        :return: Color

        """

        touch_x = (
            touch_point[0] - self.x
        )  # adjust touch position for the local position
        touch_y = screen_height - touch_point[1] - self.y

        # Call the parent's .selected function in case there is any work up there.
        # touch_point is adjusted for group's x,y position before sending to super()
        super().selected((touch_x, touch_y, 0))
        return self._color_from_position(touch_x, touch_y, self._image_size)

    def _color_from_position(self, x, y, image_size):
        img_half = image_size // 2
        dist = abs(math.sqrt((x - img_half) ** 2 + (y - img_half) ** 2))
        if x - img_half == 0:
            angle = -90
            if y > img_half:
                angle = 90
        else:
            angle = math.atan2((y - img_half), (x - img_half)) * 180 / math.pi

        angle = (angle + 30) % 360

        shade = 1 * dist / img_half
        idx = angle / 60
        base = int(round(idx))
        adj = (6 + base + (-1 if base > idx else 1)) % 6
        ratio = max(idx, base) - min(idx, base)
        color = self._make_color(base, adj, ratio, shade)

        return color

    @staticmethod
    def _make_color(base, adj, ratio, shade):
        """
        Go through each bit of the colors adjusting blue with blue, red with red,
        green with green, etc.
        """
        color_wheel = [
            [0xFF, 0x00, 0xFF],
            [0xFF, 0x00, 0x00],
            [0xFF, 0xFF, 0x00],
            [0x00, 0xFF, 0x00],
            [0x00, 0xFF, 0xFF],
            [0x00, 0x00, 0xFF],
            [0xFF, 0x00, 0xFF],
        ]

        output = 0x0
        bit = 0

        for pos in range(3):
            base_chan = color_wheel[base][pos]
            adj_chan = color_wheel[adj][pos]
            new_chan = int(round(base_chan * (1 - ratio) + adj_chan * ratio))

            # now alter the channel by the shade
            if shade < 1:
                new_chan = new_chan * shade
            elif shade > 1:
                shade_ratio = shade - 1
                new_chan = (0xFF * shade_ratio) + (new_chan * (1 - shade_ratio))

            output = output + (int(new_chan) << bit)
            bit = bit + 8
        return output



# SPDY-FileCopyrightText: 2012 jacksongabbard
# SPDX-FileCopyrightText: 2019 Dave Astels for Adafruit Industries
# SPDX-FileCopyrightText: 2021 Kevin Matocha, Jose David M.
#
# SPDX-License-Identifier: MIT
# SPDY-License-Identifier: Unlicense

"""
`wheel_maker`
================================================================================

Save a displayio.Bitmap (and associated displayio.Palette) in a BMP file.
This script is adapted in the works from Dave Astels on the ``adafruit_bitmapsaver``
and the works of  Jackson Glabbard
https://jg.gg/2012/05/28/generating-a-color-picker-style-color-wheel-in-python/
https://github.com/jacksongabbard/Python-Color-Gamut-Generator/blob/master/color-wheel-generator.py
and Kevin Matocha on the ``switch_round`` for the ``_color_to_tuple`` function

* Author(s): Dave Astels, Jackson Glabbard, Kevin Matocha, Jose David M.

Implementation Notes
--------------------



"""

import math
import struct
import gc
import board
import digitalio
import busio

try:
    import adafruit_sdcard
    import storage
except ImportError:
    pass

# pylint: disable=invalid-name, no-member, too-many-locals
# pylint: disable=consider-using-with, duplicate-code


def _write_bmp_header(output_file, filesize):
    output_file.write(bytes("BM", "ascii"))
    output_file.write(struct.pack("<I", filesize))
    output_file.write(b"\00\x00")
    output_file.write(b"\00\x00")
    output_file.write(struct.pack("<I", 54))


def _bytes_per_row(source_width):
    pixel_bytes = 3 * source_width
    padding_bytes = (4 - (pixel_bytes % 4)) % 4
    return pixel_bytes + padding_bytes


def _write_dib_header(output_file, width, height):
    output_file.write(struct.pack("<I", 40))
    output_file.write(struct.pack("<I", width))
    output_file.write(struct.pack("<I", height))
    output_file.write(struct.pack("<H", 1))
    output_file.write(struct.pack("<H", 24))
    for _ in range(24):
        output_file.write(b"\x00")


def make_color(base, adj, ratio, shade):
    """
    Go through each bit of the colors adjusting blue with blue, red with red,
    green with green, etc.
    """
    output = 0x0
    bit = 0
    for pos in range(3):
        base_chan = color_wheel[base][pos]
        adj_chan = color_wheel[adj][pos]
        new_chan = int(round(base_chan * (1 - ratio) + adj_chan * ratio))

        # now alter the channel by the shade
        if shade < 1:
            new_chan = new_chan * shade
        elif shade > 1:
            shade_ratio = shade - 1
            new_chan = (0xFF * shade_ratio) + (new_chan * (1 - shade_ratio))

        output = output + (int(new_chan) << bit)
        bit = bit + 8
    return output


def color_to_tuple(value):
    """Converts a color from a 24-bit integer to a tuple.
    :param value: RGB LED desired value - can be a RGB tuple or a 24-bit integer.
    """
    if isinstance(value, tuple):
        return value
    if isinstance(value, int):
        if value >> 24:
            raise ValueError("Only bits 0->23 valid for integer input")
        r = value >> 16
        g = (value >> 8) & 0xFF
        b = value & 0xFF
        return [r, g, b]

    raise ValueError("Color must be a tuple or 24-bit integer value.")


color_wheel = [
    [0xFF, 0x00, 0xFF],
    [0xFF, 0x00, 0x00],
    [0xFF, 0xFF, 0x00],
    [0x00, 0xFF, 0x00],
    [0x00, 0xFF, 0xFF],
    [0x00, 0x00, 0xFF],
    [0xFF, 0x00, 0xFF],
]


def make_wheel(image_name, img_size, bg_color=0x000000):
    """
    :param image_name: Name of the ouput bitmap image
    :param img_size: size of the bitmap image in pixels height=width
    :param bg_color: color of the background in 24 bit format. Defaults 0x000000
    :return: color


    **Quickstart: Importing and using the make_wheel**

    Here is one way of importing the ``make_wheel`` module so you can use:

    .. code-block:: python

        from adafruit_displayio_color_picker import wheel_maker

    Now you can create a wheel of 200 pixels with a black background using:

    .. code-block:: python

        make_wheel("wheel200.bmp", 200 , 0x000000)


    """
    img_size_width = img_size
    img_size_height = img_size
    img_half = img_size / 2
    outer_radius = img_size // 2
    background_color = color_to_tuple(bg_color)
    row_buffer = bytearray(_bytes_per_row(img_size_width))
    result_buffer = bytearray(2048)

    spi = busio.SPI(board.SCK, MOSI=board.MOSI, MISO=board.MISO)
    cs = digitalio.DigitalInOut(board.SD_CS)
    sdcard = adafruit_sdcard.SDCard(spi, cs)
    vfs = storage.VfsFat(sdcard)
    storage.mount(vfs, "/sd")
    file_path = "/sd" + image_name
    print("saving starts")
    output_file = open(file_path, "wb")
    filesize = 54 + img_size_height * _bytes_per_row(img_size_width)
    _write_bmp_header(output_file, filesize)
    _write_dib_header(output_file, img_size_width, img_size_height)

    for y in range(img_size, 0, -1):
        buffer_index = 0
        for x in range(img_size):
            dist = abs(math.sqrt((x - img_half) ** 2 + (y - img_half) ** 2))
            shade = 1 * dist / outer_radius
            if x - img_half == 0:
                angle = -90
                if y > img_half:
                    angle = 90
            else:
                angle = math.atan2((y - img_half), (x - img_half)) * 180 / math.pi

            angle = (angle - 30) % 360

            idx = angle / 60
            if idx < 0:
                idx = 6 + idx
            base = int(round(idx))

            adj = (6 + base + (-1 if base > idx else 1)) % 6
            ratio = max(idx, base) - min(idx, base)
            color = make_color(base, adj, ratio, shade)

            if dist > outer_radius:
                color_rgb = background_color
            else:
                color_rgb = color_to_tuple(color)

            for b in color_rgb:
                row_buffer[buffer_index] = b & 0xFF
                buffer_index += 1
        output_file.write(row_buffer)
        for i in range(img_size_width * 2):
            result_buffer[i] = 0
        gc.collect()

    output_file.close()
    print("saving done")
